#include <GLUT/glut.h>
#include <stdlib.h>
#include <math.h>
#include "UniformGrid.h"
#include <iostream>



float Xmin	= 0, Xmax	= 1;							//Range of variable x that we sample. This demo constrains x to [0,1]
float Ymin	= 0, Ymax	= 1;							//Range of variable y that we sample. This demo constrains y to [0,1]
int   Nx	= 100, Ny	= 100;							//Number of samples taken along the x and y axes

const int NPIX  = 1024;									//Window size, in pixels
const int NPN   = 128;									//Size of the texture pattern, in pixels
const int Npatterns = 64;								//Number of noise-texture patterns
const float dmax   = 4.0/NPIX;							//Maximum warping-factor, or displacement, allowed		
const float tmax   = 1.0/(dmax*NPN);					//Maximum warping-factor, in texture coordinates
float     alpha  = 0.12;								//Pattern injection strength (0=no patterns, 1=patterns only)
int       vector_field = 0;								//Type of vector field to visualize. This demo has two such fields.
int       iframe = 0;									//Frame-number in the animation. Used to generate time-dependent vector fields
bool      running = true;								//Flag to start/stop the animation



UniformGrid			grid(Nx,Ny,Xmin,Ymin,Xmax,Ymax);	//An uniform grid sampling our time-dependent vector field




void initGL()											//Initialize various OpenGL state variables
{ 
   glViewport(0, 0, NPIX, NPIX);
   glMatrixMode(GL_PROJECTION);							//Projection: this assumes our grid fits into the [0,1] square
   glLoadIdentity(); 
   glTranslatef(-1.0, -1.0, 0.0); 
   glScalef(2.0, 2.0, 1.0);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT); 
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT); 
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
   glEnable(GL_TEXTURE_2D);
   glShadeModel(GL_FLAT);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glClear(GL_COLOR_BUFFER_BIT);
}


void makeNoisePatterns()								//Creates the Npattern noise-texture patterns, and stores them
{														//into separate display-lists. This way, we can later on easily
   int lut[256];										//use each such pattern, just by calling the respective display list.	
   int phase[NPN][NPN];
   GLubyte pat[NPN][NPN][4];
   int i, j, k, t;
    
   for (i = 0; i < 256; i++)							//Create the luminance colormap used for noise patterns. 
      lut[i] = i < 127 ? 0 : 255;						//Here, we use a simple two-color (black-and-white) colormap.
														//Try different other colormaps!
	  
   for (i = 0; i < NPN; i++)							//Create the random phase pattern
	for (j = 0; j < NPN; j++) 
	  phase[i][j] = rand() % 256; 

   for (k = 0; k < Npatterns; k++)						//Create the Npatterns texture-patterns:
   {
      t = k*256/Npatterns;
      for (i = 0; i < NPN; i++) 
        for (j = 0; j < NPN; j++) 
	    {
          pat[i][j][0] =								//Pattern luminance: color-mapped phase
          pat[i][j][1] =
          pat[i][j][2] = lut[(t + phase[i][j]) % 255];
          pat[i][j][3] = 255*alpha;						//Pattern alpha: tells how 'strongly' this pattern will be injected
        }
      glNewList(k+1, GL_COMPILE);						//Save the texture as a display list for further use
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, NPN, NPN, 0, GL_RGBA, GL_UNSIGNED_BYTE, pat);
      glEndList();
   }
}




float f(float x, float y)								//A simple two-variable function. Used to construct vector-field 0
{														//(see computeVectors0)
    float xc = (Xmax+Xmin)/2, yc = (Ymax+Ymin)/2;
	float ret = 8*expf(-((x-xc)*(x-xc)+(y-yc)*(y-yc))/5);
	return ret;
}




void computeVectors0(UniformGrid& grid)					//Create vectors for the grid. For this, we simply sample the 
{														//rotated gradient of the function f(x,y) at all the grid points.
    float rotation = iframe/100.0f;						//To create a changing field, we make the rotation time-dependent.

	for(int i=0;i<grid.numPoints();++i)
	{
	   float p[2];				
	   grid.getPoint(i,p);								//Get coordinates of current grid point
	   
	   float scalar = f(p[0],p[1]);						//Evaluate f(x,y) at current point
		
	   float delta    = 0.001*(Xmax-Xmin);				//Evaluate gradient of f(x,y) using finite differences
	   float scalar_x = f(p[0]+delta,p[1]);	
	   float scalar_y = f(p[0],p[1]+delta);		   
	   Point3d gradient = Point3d((scalar_x-scalar)/delta,(scalar_y-scalar)/delta,0);
	   
	   Point3d vec;										//Rotate the gradient with an angle alpha, to generate more complex vector fields
	   vec.x = cos(rotation)*gradient.x - sin(rotation)*gradient.y;
	   vec.y = sin(rotation)*gradient.x + cos(rotation)*gradient.y;
	   
	   float length = vec.norm();						//Make sure that the vector length is limited by 'dmax'	
	   vec *= dmax/length;
	   	   
	   grid.pointVectors().setC0Vector(i,vec);			//Finally, store the rotated gradient as vector values in the grid
	}   
}



void computeVectors1(UniformGrid& grid)					//Initialize the point-vectors of 'grid' with a second time-dependent vector field
{
   VectorAttributes& vectors = grid.pointVectors();

   float  sa  = 0.010*cos(iframe*2.0*M_PI/200.0);		//Time is computed as the current frame in the animation

   float sizeX = Xmax-Xmin, sizeY = Ymax-Ymin;
   
   for(int i=0;i<grid.numPoints();++i)
   {
        float p[2],v[2];
	    grid.getPoint(i,p);

		float dx = p[0] - sizeX/2;         
		float dy = p[1] - sizeY/2; 
		float r  = dx*dx + dy*dy; 
		if (r < 0.0001) r = 0.0001;
		v[0] = sa*dx/r + 0.02;  
		v[1] = sa*dy/r;
		r  = v[0]*v[0]+v[1]*v[1];
		if (r > dmax*dmax)								//Make sure that the vector length is limited by 'dmax'
		{ 
		  r  = sqrt(r); 
		  v[0] *= dmax/r; 
		  v[1] *= dmax/r; 
		  vectors.setC0Vector(i,v);
		}
   }
}




void warpPoint(float* p,UniformGrid& grid,int index)	//Warp a grid point by moving it along its vector data	
{
   VectorAttributes& v = grid.pointVectors();
   
   float* vec = v.getC0Vector(index);

   p[0] += vec[0];
   p[1] += vec[1];
}



void display()											//Central IBFV function: This constructs and displays a new animation frame
{ 
   if (vector_field==0)									//1. Compute the new values of the time-dependent vector field on 'grid'
	  computeVectors0(grid);	
   else 
      computeVectors1(grid);	  

   for (int i = 0; i<grid.numCells(); i++)				//2. Warp the grid using the vector field. Also, draw the warped grid,
   {													//   textured by the image from the last animation frame.
      int cell[4];
	  grid.getCell(i,cell);
	  
	  float p0[2],p1[2],p2[2],p3[2];
	  grid.getPoint(cell[0],p0);
	  grid.getPoint(cell[1],p1);
	  grid.getPoint(cell[2],p2);
	  grid.getPoint(cell[3],p3);
	  
      glBegin(GL_QUADS);
	  glTexCoord2f(p0[0],p0[1]); warpPoint(p0,grid,cell[0]); glVertex2f(p0[0],p0[1]);
	  glTexCoord2f(p1[0],p1[1]); warpPoint(p1,grid,cell[1]); glVertex2f(p1[0],p1[1]);
	  glTexCoord2f(p2[0],p2[1]); warpPoint(p2,grid,cell[2]); glVertex2f(p2[0],p2[1]);
	  glTexCoord2f(p3[0],p3[1]); warpPoint(p3,grid,cell[3]); glVertex2f(p3[0],p3[1]);
      glEnd();
   }
   
   ++iframe;											//3. Advance the frame counter

   glEnable(GL_BLEND);									//4. Inject the next noise pattern in our sequence of Npatterns.
   glCallList(iframe % Npatterns + 1);					//   For this, we draw a single quad covering the entire window, 
   glBegin(GL_QUAD_STRIP);								//   textured by this pattern
      glTexCoord2f(0.0,  0.0);  glVertex2f(0.0, 0.0);
      glTexCoord2f(0.0,  tmax); glVertex2f(0.0, 1.0);
      glTexCoord2f(tmax, 0.0);  glVertex2f(1.0, 0.0);
      glTexCoord2f(tmax, tmax); glVertex2f(1.0, 1.0);
   glEnd();
   glDisable(GL_BLEND);

   glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 0, 0, NPIX, NPIX, 0);
														//5. Copy the current image. We'll use this as the texture to be warped
														//   at step (2) in the next animation frame.	

   glutSwapBuffers(); 
}


void keyboard(unsigned char c,int,int)					//Keyboard callback. Controls various animation aspects:
{
	switch (c)
	{
	case 'a':											// 'a': Decrease the noise-injection strength
		alpha /= 1.1; 
		if (alpha<0.02) alpha = 0.02; 
		makeNoisePatterns(); 
		break;
	case 'A':											// 'A': Decrease the noise-injection strength	
		alpha *= 1.1;
		if (alpha>1) alpha = 1; 
		makeNoisePatterns(); 
		break;
	case 'v':											// 'v': Toggle between the two vector fields we can show
	    vector_field = (++vector_field) % 2;	
		break;
	case ' ':
		running = !running;
		glutIdleFunc((running)? display:0);
		break; 
	}
	
	glutPostRedisplay();
}



int main(int argc, char** argv)							//Main program
{ 
   cout<<"Image based flow visualization (IBFV)"<<endl;
   cout<<"Visualization controls:"<<endl;
   cout<<"   -a,A:    decrease/increase the noise injection strength"<<endl;
   cout<<"   -v:      toggle between different vector fields to visualize"<<endl;
   cout<<"   -space:  start/stop the animation"<<endl;	

   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB); 
   glutInitWindowSize(NPIX, NPIX);
   glutCreateWindow(argv[0]);
   initGL();

   glutDisplayFunc(display);
   glutIdleFunc(display);
   glutKeyboardFunc(keyboard);
   
   makeNoisePatterns();
   glutMainLoop();
   return 0;
}

